import scipy.special as sp
import numpy as np
from mielib import extraspecial


def optics_mie_a(n, k0a, eps_p, mu_p=1, eps_h=1, mu_h=1, get_denominator=False):
    """
        Electric Mie coefficent. For detatails see Bohren p. 100

        Arguments:
            n - 2^n multipole order
            k0a - vacuum size parameter
            eps_p, mu_p - particle parameters
            esp_h, mu_h - host parameters
    """
    n_p, n_h = np.sqrt(eps_p * mu_p, dtype=complex), np.sqrt(eps_h * mu_h, dtype=complex)
    x = n_h * k0a
    m = n_p / n_h
    mu = mu_p / mu_h

    mx = m * x
    jnmx = sp.spherical_jn(n, mx)
    jnx = sp.spherical_jn(n, x)
    h1nx = extraspecial.spherical_h1(n, x)
    xjnx_p = jnx + x * sp.spherical_jn(n, x, 1)
    mxjnmx_p = jnmx + mx * sp.spherical_jn(n, mx, 1)
    xh1nx_p = h1nx + x * extraspecial.spherical_h1(n, x, p=1)

    if get_denominator:
        # for dispersion of TM modes
        return m**2 * jnmx * xh1nx_p - mu * h1nx * mxjnmx_p
    else:
        return (m**2 * jnmx * xjnx_p - mu * jnx * mxjnmx_p) / (m**2 * jnmx * xh1nx_p - mu * h1nx * mxjnmx_p)


def optics_mie_b(n, k0a, eps_p, mu_p=1, eps_h=1, mu_h=1, get_denominator=False):
    """
        Electric Mie coefficent. For detatails see Bohren p. 100

        Arguments:
            n - 2^n multipole order
            k0a - vacuum size parameter
            eps_p, mu_p - particle parameters
            esp_h, mu_h - host parameters
    """
    n_p, n_h = np.sqrt(eps_p * mu_p, dtype=complex), np.sqrt(eps_h * mu_h, dtype=complex)
    x = n_h * k0a
    m = n_p / n_h
    mu = mu_p / mu_h

    mx = m * x
    jnmx = sp.spherical_jn(n, mx)
    jnx = sp.spherical_jn(n, x)
    h1nx = extraspecial.spherical_h1(n, x)
    xjnx_p = jnx + x * sp.spherical_jn(n, x, 1)
    mxjnmx_p = jnmx + mx * sp.spherical_jn(n, mx, 1)
    xh1nx_p = h1nx + x * extraspecial.spherical_h1(n, x, p=1)
    
    if get_denominator:
        # for dispertion of TE modes
        return mu * jnmx * xh1nx_p - h1nx * mxjnmx_p
    else:
        return (mu * jnmx * xjnx_p - jnx * mxjnmx_p) / (mu * jnmx * xh1nx_p - h1nx * mxjnmx_p)


def optics_scattering_cross_section(k0, a, eps_p, mu_p=1, eps_h=1, mu_h=1, nmin=1, nmax=50, norm='none'):
    k0a = np.asarray(k0 * a)

    n_host = np.sqrt(eps_h * mu_h)

    sigma_norm = 1.0
    if norm == 'geom':
        sigma_norm = np.pi * a**2

    sigma_sc   = np.zeros(k0a.size, dtype=np.float64)
    sigma_sc_n_electric = np.zeros([nmax, k0a.size], dtype=np.float64)
    sigma_sc_n_magnetic = np.zeros([nmax, k0a.size], dtype=np.float64)

    for n in range(nmin, nmax):
        an = optics_mie_a(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        bn = optics_mie_b(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        sigma_sc_n_electric[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * np.abs(an**2)
        sigma_sc_n_magnetic[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * np.abs(bn**2)
        
    sigma_sc = np.sum(sigma_sc_n_electric + sigma_sc_n_magnetic, axis=0)

    return sigma_sc/sigma_norm, sigma_sc_n_electric/sigma_norm, sigma_sc_n_magnetic/sigma_norm



def optics_extinction_cross_section(k0, a, eps_p, mu_p=1, eps_h=1, mu_h=1, nmin=1, nmax=50, norm='none'):
    k0a = np.asarray(k0 * a)

    n_host = np.sqrt(eps_h * mu_h)

    sigma_norm = 1.0
    if norm == 'geom':
        sigma_norm = np.pi * a**2

    sigma_ext   = np.zeros(k0a.size, dtype=np.float64)
    sigma_ext_n_electric = np.zeros([nmax, k0a.size], dtype=np.float64)
    sigma_ext_n_magnetic = np.zeros([nmax, k0a.size], dtype=np.float64)

    for n in range(nmin, nmax):
        an = optics_mie_a(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        bn = optics_mie_b(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        sigma_ext_n_electric[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * np.real(an)
        sigma_ext_n_magnetic[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * np.real(bn)
        
    sigma_ext = np.sum(sigma_ext_n_electric + sigma_ext_n_magnetic, axis=0)

    return sigma_ext/sigma_norm, sigma_ext_n_electric/sigma_norm, sigma_ext_n_magnetic/sigma_norm


def optics_absorption_cross_section(k0, a, eps_p, mu_p=1, eps_h=1, mu_h=1, nmin=1, nmax=50, norm='none'):
    k0a = np.asarray(k0 * a)

    n_host = np.sqrt(eps_h * mu_h)

    sigma_norm = 1.0
    if norm == 'geom':
        sigma_norm = np.pi * a**2

    sigma_abs   = np.zeros(k0a.size, dtype=np.float64)
    sigma_abs_n_electric = np.zeros([nmax, k0a.size], dtype=np.float64)
    sigma_abs_n_magnetic = np.zeros([nmax, k0a.size], dtype=np.float64)

    for n in range(nmin, nmax):
        an = optics_mie_a(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        bn = optics_mie_b(n, k0a, eps_p=eps_p, mu_p=mu_p, eps_h=eps_h, mu_h=mu_h)
        sigma_abs_n_electric[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * (np.real(an) - np.abs(an)**2)
        sigma_abs_n_magnetic[n, :] = 2*np.pi / (n_host * k0)**2 * (2*n+1) * (np.real(bn) - np.abs(bn)**2)
        
    sigma_abs = np.sum(sigma_abs_n_electric + sigma_abs_n_magnetic, axis=0)

    return sigma_abs/sigma_norm, sigma_abs_n_electric/sigma_norm, sigma_abs_n_magnetic/sigma_norm


import numpy as np

def rotation_matrix_3d_sph(theta, phi):
    """
        A^cart = R(theta, phi) A^sph

        https://en.wikipedia.org/wiki/Vector_fields_in_cylindrical_and_spherical_coordinates
    """
    return np.array([
        [np.sin(theta)*np.cos(phi), np.cos(theta)*np.cos(phi), -np.sin(phi)],
        [np.sin(theta)*np.sin(phi), np.cos(theta)*np.sin(phi),  np.cos(phi)],
        [            np.cos(theta),            -np.sin(theta),            0]
    ])


def rotation_matrix_2d(phi):
    """
        A^cart = R(phi) A^cyl

        https://en.wikipedia.org/wiki/Vector_fields_in_cylindrical_and_spherical_coordinates
    """
    return np.array([
        [np.cos(phi), -np.sin(phi)],
        [np.sin(phi),  np.cos(phi)]
    ])


def rotation_matrix_3d_cyl(phi, axis='z'):
    """
        A^cart = R(phi) A^cyl

        https://en.wikipedia.org/wiki/Vector_fields_in_cylindrical_and_spherical_coordinates
    """
    match axis:
        case 'x':
            return np.nan
        case 'y':
            return np.nan
        case 'z':
            return np.array([
                [np.cos(phi), -np.sin(phi), 0],
                [np.sin(phi),  np.cos(phi), 0],
                [          0,            0, 1],
            ])
        case _:
            return np.eye(3)


def cart2sph(x, y, z):
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z/r)
    phi = np.arctan2(y, x)

    return r, theta, phi


def sph2cart(r, theta, phi):
    x = r * np.sin(theta) * np.cos(phi)
    x = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    return x, y, z


def cart2cyl(x, y, z):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)

    return rho, phi, z


def cyl2cart(rho, phi, z):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)

    return x, y, z


import numpy as np
import scipy.special as sp

from mielib import extraspecial

def spherical_radial_function_z(j, rho, p=0, superscript=1):
    """
        j - radial quantum number
        rho - argument
        p = 0,1 - derivative order
        superscript - type of radial function
    """
    zj = np.array([0.0j])
    match superscript:
        case 1:
            zj = sp.spherical_jn(j, rho, p)
        case 2:
            zj = sp.spherical_yn(j, rho, p)
        case 3:
            zj = extraspecial.spherical_h1(j, rho, p)
        case 4:
            zj = extraspecial.spherical_h2(j, rho, p)

    return zj

        
def vsh_bohren_me(m, n, rho, theta, phi, superscript):
    '''
        vector Harmonics in spherical coordinates (r, θ, φ)
        from Bohren & Huffman, eq. (4.17), (4.18)
        Input
        -----
            rho = k r, k = k0 n_media
            theta = [0, pi] -- azimuthal angle
            phi = [0, 2pi] -- polar angle
            n = 0, 1, ...
            m = 0, ..., n
            superscript = 1, 2, 3, 4
    '''
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    Mt = 0.0
    Mp = 0.0

    zn = spherical_radial_function_z(n, rho, p=0, superscript=superscript)

    Mt = -m / np.sin(theta) * np.sin(m * phi) * sp.lpmv(m, n, np.cos(theta))

    # formula for the d_θ Pnm(cos θ) is from here (I got a different sign though...)
    # https://math.stackexchange.com/questions/391672/derivative-of-associated-legendre-polynomials-at-x-pm-1
    # https://mathworld.wolfram.com/AssociatedLegendrePolynomial.html
    dPnm = 1/np.sin(theta) * (np.abs(np.sin(theta)) * sp.lpmv(m+1, n,
                                                              np.cos(theta)) + m * np.cos(theta) * sp.lpmv(m, n, np.cos(theta)))

    # an old way
    # dtheta = 1e-12
    # dPnm = (lpmv(m, n, np.cos(theta + dtheta)) - lpmv(m, n, np.cos(theta - dtheta))) / (2*dtheta)

    Mp = -np.cos(m * phi) * dPnm

    Mr = np.zeros(np.shape(Mp*zn))

    return np.array([Mr, Mt*zn, Mp*zn])


def vsh_bohren_mo(m, n, rho, theta, phi, superscript):
    '''
        vector Harmonics in spherical coordinates (r, θ, φ)
        from Bohren & Huffman, eq. (4.17), (4.18)
        Input
        -----
            rho = k r, k = k0 n_media
            theta = [0, pi] -- azimuthal angle
            phi = [0, 2pi] -- polar angle
            n = 0, 1, ...
            m = 0, ..., n
            superscript = 1, 2, 3, 4
    '''
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    Mt = 0.0
    Mp = 0.0

    zn = spherical_radial_function_z(n, rho, p=0, superscript=superscript)

    Mt = m / np.sin(theta) * np.cos(m * phi) * sp.lpmv(m, n, np.cos(theta))

    # formula for the d_θ Pnm(cos θ) is from here (I got a different sign though...)
    # https://math.stackexchange.com/questions/391672/derivative-of-associated-legendre-polynomials-at-x-pm-1
    # https://mathworld.wolfram.com/AssociatedLegendrePolynomial.html
    dPnm = 1/np.sin(theta) * (np.abs(np.sin(theta)) * sp.lpmv(m+1, n,
                                                              np.cos(theta)) + m * np.cos(theta) * sp.lpmv(m, n, np.cos(theta)))

    # an old way
    # dtheta = 1e-12
    # dPnm = (lpmv(m, n, np.cos(theta + dtheta)) - lpmv(m, n, np.cos(theta - dtheta))) / (2*dtheta)

    Mp = -np.sin(m * phi) * dPnm

    Mr = np.zeros(np.shape(Mp*zn))

    return np.array([Mr, Mt*zn, Mp*zn])


def vsh_bohren_ne(m, n, rho, theta, phi, superscript):
    '''
        vector Harmonics in spherical coordinates (r, θ, φ)
        from Bohren & Huffman, eq. (4.19), (4.20)
        Input
        -----
            rho = k r, k = k0 n_media
            theta = [0, pi] -- azimuthal angle
            phi = [0, 2pi] -- polar angle
            n = 0, 1, ...
            m = 0, ..., n
            superscript = 1, 2, 3, 4
    '''
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    Nr = 0.0
    Nt = 0.0
    Np = 0.0

    zn = spherical_radial_function_z(n, rho, p=0, superscript=superscript)
    znp = spherical_radial_function_z(n, rho, p=1, superscript=superscript)

    Pnm = sp.lpmv(m, n, np.cos(theta))

    # formula for the d_θ Pnm(cos θ) is from here (I got a different sign though...)
    # https://math.stackexchange.com/questions/391672/derivative-of-associated-legendre-polynomials-at-x-pm-1
    # https://mathworld.wolfram.com/AssociatedLegendrePolynomial.html
    dPnm = 1/np.sin(theta) * (np.abs(np.sin(theta)) * sp.lpmv(m+1, n,
                                                              np.cos(theta)) + m * np.cos(theta) * sp.lpmv(m, n, np.cos(theta)))

    # an old way
    # dtheta = 1e-12
    # dPnm = (lpmv(m, n, np.cos(theta + dtheta)) - lpmv(m, n, np.cos(theta - dtheta))) / (2*dtheta)

    Nr = zn/rho * np.cos(m*phi) * n * (n+1) * Pnm
    Nt = np.cos(m * phi) * dPnm * (zn/rho + znp)
    Np = -m * np.sin(m * phi) * Pnm / np.sin(theta) * (zn/rho + znp)

    return np.array([Nr, Nt, Np])


def vsh_bohren_no(m, n, rho, theta, phi, superscript=1):
    '''
        vector Harmonics in spherical coordinates (r, θ, φ)
        from Bohren & Huffman, eq. (4.19), (4.20)
        Input
        -----
            rho = k r, k = k0 n_media
            theta = [0, pi] -- azimuthal angle
            phi = [0, 2pi] -- polar angle
            n = 0, 1, ...
            m = 0, ..., n
            superscript = 1, 2, 3, 4
    '''
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    Nr = 0.0
    Nt = 0.0
    Np = 0.0

    zn = spherical_radial_function_z(n, rho, p=0, superscript=superscript)
    znp = spherical_radial_function_z(n, rho, p=1, superscript=superscript)

    Pnm = sp.lpmv(m, n, np.cos(theta))

    # formula for the d_θ Pnm(cos θ) is from here (I got a different sign though...)
    # https://math.stackexchange.com/questions/391672/derivative-of-associated-legendre-polynomials-at-x-pm-1
    # https://mathworld.wolfram.com/AssociatedLegendrePolynomial.html
    dPnm = 1/np.sin(theta) * (np.abs(np.sin(theta)) * sp.lpmv(m+1, n,
                                                              np.cos(theta)) + m * np.cos(theta) * sp.lpmv(m, n, np.cos(theta)))

    # an old way
    # dtheta = 1e-12
    # dPnm = (lpmv(m, n, np.cos(theta + dtheta)) - lpmv(m, n, np.cos(theta - dtheta))) / (2*dtheta)

    Nr = zn/rho * np.sin(m * phi) * n * (n+1) * Pnm
    Nt = np.sin(m * phi) * dPnm * (zn/rho + znp)
    Np = m * np.cos(m * phi) * Pnm / np.sin(theta) * (zn/rho + znp)

    return np.array([Nr, Nt, Np])


def vsh_jackson_x(m, n, theta, phi):
    '''
        from Jackson's book paragraph 9.7
    '''
    Ymn = sp.sph_harm(m, n, phi, theta)
    dYmn = m/np.tan(theta) * Ymn
    if m+1 <= n:
        # sqrt((-m + n) (1 + m + n)) = np.sqrt(sp.gamma(1-m+n)) * np.sqrt(sp.gamma(2+m+n)) / (np.sqrt(sp.gamma(-m+n)) * np.sqrt(sp.gamma(1+m+n)))
        dYmn += np.exp(-1j*phi) * np.sqrt((-m + n)*(1 + m + n)
                                          ) * sp.sph_harm(m+1, n, phi, theta)

    return (-1)/np.sqrt(n*(n+1)) * np.array([
        np.zeros(np.shape(theta * phi)),
        m/np.sin(theta) * Ymn,
        1j * dYmn
    ])


def vsh_toftul_m(m, j, rho, theta, phi, superscript=3):
    """ 
        M vector spherical harmonic

        Arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
    """
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    zj = spherical_radial_function_z(j, rho, p=0, superscript=superscript)

    Mt = 1j * m / np.sin(theta) * zj * sp.sph_harm(m, j, phi, theta)

    dYmn = m/np.tan(theta) * sp.sph_harm(m, j, phi, theta)
    if m+1 <= j:
        # sqrt((-m + j) (1 + m + j)) = np.sqrt(sp.gamma(1-m+j)) * np.sqrt(sp.gamma(2+m+j)) / (np.sqrt(sp.gamma(-m+j)) * np.sqrt(sp.gamma(1+m+j)))
        dYmn += np.exp(-1j*phi) * np.sqrt((-m + j)*(1 + m + j)) * sp.sph_harm(m+1, j, phi, theta)

    Mp = -zj * dYmn

    Mr = np.zeros(np.shape(Mp))

    return np.array([Mr, Mt, Mp])


def vsh_toftul_n(m, j, rho, theta, phi, superscript=3):
    """ 
        N vector spherical harmonic

        Arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
    """
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    zj  = spherical_radial_function_z(j, rho, p=0, superscript=superscript)
    zjp = spherical_radial_function_z(j, rho, p=1, superscript=superscript)

    Ymn = sp.sph_harm(m, j, phi, theta)
    dYmn = m/np.tan(theta) * Ymn
    if m+1 <= j:
        # sqrt((-m + j) (1 + m + j)) = np.sqrt(sp.gamma(1-m+j)) * np.sqrt(sp.gamma(2+m+j)) / (np.sqrt(sp.gamma(-m+j)) * np.sqrt(sp.gamma(1+m+j)))
        dYmn += np.exp(-1j*phi) * np.sqrt((-m + j)*(1 + m + j)) * sp.sph_harm(m+1, j, phi, theta)

    Nr = j*(j+1) * zj/rho * Ymn
    Nt = 1/rho * (zj + rho * zjp) * dYmn
    Np = 1j*m/np.sin(theta) * 1/rho * (zj + rho * zjp) * Ymn

    return np.array([Nr, Nt, Np])


def vsh_toftul_l(m, j, rho, theta, phi, superscript=3):
    """ 
        L vector spherical harmonic

        Arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
    """
    # convert input to np arrays
    rho = np.asarray(rho, dtype=complex)
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    # to prevent devision by zero
    rho[np.abs(rho) < 1e-15] = 1e-15
    theta[np.abs(theta) < 1e-15] = 1e-15

    zj  = spherical_radial_function_z(j, rho, p=0, superscript=superscript)
    zjp = spherical_radial_function_z(j, rho, p=1, superscript=superscript)

    Ymn = sp.sph_harm(m, j, phi, theta)
    dYmn = m/np.tan(theta) * Ymn
    if m+1 <= j:
        # sqrt((-m + n) (1 + m + n)) = np.sqrt(sp.gamma(1-m+n)) * np.sqrt(sp.gamma(2+m+n)) / (np.sqrt(sp.gamma(-m+n)) * np.sqrt(sp.gamma(1+m+n)))
        dYmn += np.exp(-1j*phi) * np.sqrt((-m + j)*(1 + m + j)) * sp.sph_harm(m+1, j, phi, theta)

    Lr = zjp * Ymn
    Lt = zj/rho * dYmn
    Lp = 1j*m/np.sin(theta) * zj/rho * Ymn

    return np.array([Lr, Lt, Lp])


# wrapper functions
def vector_spherical_harmonic_m(m, j, rho, theta, phi, superscript=3, source='toftul', parity='even'):
    """ 
        vector spherical harmonic

        arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
            source:
                'toftul' - based on sm arxiv.org/abs/2210.04021
                'bohren' - based on bohren & huffmann book
            partiy (applicapble only for real vsh):
                'even' - even harmonic
                'odd' - odd harmonic
    """
    match source:
        case 'toftul':
            return vsh_toftul_m(m, j, rho, theta, phi, superscript=superscript)
        case 'bohren':
            match parity:
                case 'even':
                    return vsh_bohren_me(m=m, n=j, rho=rho, theta=theta, phi=phi, superscript=superscript)
                case 'odd':
                    return vsh_bohren_mo(m=m, n=j, rho=rho, theta=theta, phi=phi, superscript=superscript)
                case _:
                    return np.array([0.0, 0.0, 0.0])
        case _:
            return np.array([0.0, 0.0, 0.0])
     


def vector_spherical_harmonic_n(m, j, rho, theta, phi, superscript=3, source='toftul', parity='even'):
    """ 
        vector spherical harmonic

        arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
            source:
                'toftul' - based on sm arxiv.org/abs/2210.04021
                'bohren' - based on bohren & huffmann book
            partiy (applicapble only for real vsh):
                'even' - even harmonic
                'odd' - odd harmonic
    """
    match source:
        case 'toftul':
            return vsh_toftul_n(m, j, rho, theta, phi, superscript=superscript)
        case _:
            return np.array([0.0, 0.0, 0.0])


def vector_spherical_harmonic_l(m, j, rho, theta, phi, superscript=3, source='toftul', parity='even'):
    """ 
        vector spherical harmonic

        arguments:
            m - projection of total angular momentum 
            j - total angular momentum
            rho, theta, phi - arguments in spherical coordinate system
            superscript:
                1 - spherical bessel
                3 - spherical hankel1
            source:
                'toftul' - based on sm arxiv.org/abs/2210.04021
                'bohren' - based on bohren & huffmann book
            partiy (applicapble only for real vsh):
                'even' - even harmonic
                'odd' - odd harmonic
    """
    match source:
        case 'toftul':
            return vsh_toftul_l(m, j, rho, theta, phi, superscript=superscript)
        case _:
            return np.array([0.0, 0.0, 0.0])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mielib']

package_data = \
{'': ['*']}

install_requires = \
['scipy>=1.9.3,<2.0.0']

setup_kwargs = {
    'name': 'mielib',
    'version': '0.1.10',
    'description': '',
    'long_description': '# MieLib\n\nPython library with many Mie-related functions for optics and acoustics. In particular:\n- Scattering Mie coefficients for isotropic spheres:\n    -  Acoustic Mie coefficients based on Phys. Rev. Lett. 123, 183901 (2019) (see SM)\n    - Optics Mie coefficients based on Bohren Huffmann book.\n- Scalar spherical harmonics\n- Vector spherical harmonics (complex and real)\n\n## Instalation\n\nFrom [PYPI/mielib](https://pypi.org/project/mielib/):\n```\npip install mielib\n```\n\n## Examples\n\nSee examples folder with Jupyter Notebooks.\n\n## ToDo\n\n* Make tests\n* Setup github actions\n* Search for complex poles of Mie scattering coefficients\n\n## Credits\nIvan Toftul \n\n`toftul.ivan@gmail.com`\n',
    'author': 'Ivan Toftul',
    'author_email': 'toftul.ivan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
